# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["build_backbone"]


def build_backbone(config, model_type):
    if model_type == "det":
        from .det_mobilenet_v3 import MobileNetV3
        from .det_resnet_vd import ResNet
        from .det_resnet_vd_sast import ResNet_SAST
        from .det_vgg16_bn import VGG16_BN

        support_dict = ["VGG16_BN", "MobileNetV3", "ResNet", "ResNet_SAST"]
    elif model_type == "rec" or model_type == "cls":
        from .rec_mobilenet_v3 import MobileNetV3
        from .rec_resnet_fpn import ResNetFPN
        from .rec_resnet_vd import ResNet

        support_dict = ["MobileNetV3", "ResNet", "ResNetFPN"]
    else:
        raise NotImplementedError

    module_name = config.pop("name")
    assert module_name in support_dict, Exception(
        "when model typs is {}, backbone only support {}".format(
            model_type, support_dict
        )
    )
    module_class = eval(module_name)(**config)
    return module_class



# Copyright (c) 2020 FSMLP Authors. All Rights Reserved.
from __future__ import division, print_function

import math

import cv2
import numpy as np
import torch


class CRAFTPostProcess(object):
    def __init__(
        self,
        text_thresh=0.35,
        link_thresh=0.1,
        min_size=3,
        use_dilate=False,
        xpad=4,
        ypad=2,
        xdilate=9,
        ydilate=3,
        merged_dilation_kernel=None,
        rotated_box=True,
        **kwargs
    ):
        self.max_candidates = 1000
        self.text_thresh = text_thresh
        self.link_thresh = link_thresh
        self.min_size = min_size
        self.use_dilate = use_dilate
        self.xdilate = xdilate
        self.ydilate = ydilate
        self.xpad = xpad
        self.ypad = ypad
        self.merged_dilation_kernel = (
            None
            if merged_dilation_kernel is None
            else np.array([[1, 1], [1, 1]])
        )
        self.rotated_box = rotated_box

    def boxes_from_bitmap0(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape
        contours = None

        outs = cv2.findContours(
            (bitmap * 255).astype(np.uint8),
            cv2.RETR_LIST,
            cv2.CHAIN_APPROX_SIMPLE,
        )
        if len(outs) == 3:
            _, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)

        boxes = []
        scores = []
        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            # score = self.box_score_fast(pred, points.reshape(-1, 2))

            # if self.box_thresh > score:
            #     continue

            box = self.unclip(points).reshape(-1, 1, 2)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)

            box[:, 0] = np.clip(
                np.round(box[:, 0] / width * dest_width), 0, dest_width
            )
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height
            )
            boxes.append(box.astype(np.int16))
            # scores.append(score)

        return np.array(boxes, dtype=np.int16), scores

    def boxes_from_bitmap1(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape

        outs = cv2.findContours(
            (bitmap * 255).astype(np.uint8),
            cv2.RETR_LIST,
            cv2.CHAIN_APPROX_SIMPLE,
        )
        if len(outs) == 3:
            _, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)

        xpad = self.xpad
        ypad = self.ypad
        padding = np.array(
            [
                [-xpad, int(ypad * -1.5)],
                [int(xpad * 0.8), int(ypad * -1.5)],
                [int(xpad * 0.8), int(ypad * 0.8)],
                [-xpad, int(ypad * 0.8)],
            ]
        )
        boxes = []
        scores = []
        for index in range(num_contours):
            contour = contours[index]

            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            box = np.array(points)

            if not self.rotated_box:
                l, t, w, h = cv2.boundingRect(box)
                box = np.array(
                    [[l, t], [l + w, t], [l + w, t + h], [l, t + h]]
                )

            boxes.append(box)

        if len(boxes) == 0:
            return [], []

        boxes = np.array(boxes)

        boxes = np.add(boxes, padding)
        boxes[boxes[:, :, 0] < 0] = 0
        boxes[boxes[:, 1, 0] > width] = width
        boxes[boxes[:, 2, 0] > width] = width
        boxes[boxes[:, 3, 1] > height] = height
        boxes[boxes[:, 3, 1] > height] = height

        boxes[:, :, 0] = np.clip(
            np.round(boxes[:, :, 0] / width * dest_width), 0, dest_width
        )
        boxes[:, :, 1] = np.clip(
            np.round(boxes[:, :, 1] / height * dest_height), 0, dest_height
        )
        return boxes, scores

    def boxes_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape

        nLabels, labels, stats, centroids = cv2.connectedComponentsWithStats(
            bitmap.astype(np.uint8), connectivity=4
        )

        boxes = []
        scores = []
        for k in range(1, nLabels):
            # size filtering
            size = stats[k, cv2.CC_STAT_AREA]
            if size < 10:
                continue

            # make segmentation map
            segmap = np.zeros(bitmap.shape, dtype=np.uint8)
            segmap[labels == k] = 255  # remove link area
            x, y = stats[k, cv2.CC_STAT_LEFT], stats[k, cv2.CC_STAT_TOP]
            w, h = stats[k, cv2.CC_STAT_WIDTH], stats[k, cv2.CC_STAT_HEIGHT]
            niter = int(math.sqrt(size * min(w, h) / (w * h)) * 2)
            sx, ex, sy, ey = (
                x - niter,
                x + w + niter + 1,
                y - niter,
                y + h + niter + 1,
            )
            # boundary check
            if sx < 0:
                sx = 0
            if sy < 0:
                sy = 0
            if ex >= width:
                ex = width
            if ey >= height:
                ey = height
            kernel = cv2.getStructuringElement(
                cv2.MORPH_RECT, (1 + niter, 1 + niter)
            )
            segmap[sy:ey, sx:ex] = cv2.dilate(segmap[sy:ey, sx:ex], kernel)

            # make box
            contour = (
                np.roll(np.array(np.where(segmap != 0)), 1, axis=0)
                .transpose()
                .reshape(-1, 2)
            )
            rectangle = cv2.minAreaRect(contour)
            box = cv2.boxPoints(rectangle)

            # align diamond-shape
            w, h = (
                np.linalg.norm(box[0] - box[1]),
                np.linalg.norm(box[1] - box[2]),
            )
            box_ratio = max(w, h) / (min(w, h) + 1e-5)
            if abs(1 - box_ratio) <= 0.1:
                l, r = min(contour[:, 0]), max(contour[:, 0])
                t, b = min(contour[:, 1]), max(contour[:, 1])
                box = np.array(
                    [[l, t], [r, t], [r, b], [l, b]], dtype=np.float32
                )

            # make clock-wise order
            startidx = box.sum(axis=1).argmin()
            box = np.roll(box, 4 - startidx, 0)
            box = np.array(box)

            box[:, 0] = np.clip(
                np.round(box[:, 0] / width * dest_width), 0, dest_width
            )
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height
            )

            boxes.append(box)

        return boxes, scores

    def unclip(self, box):
        import pyclipper
        from shapely.geometry import Polygon

        unclip_ratio = 1.1
        poly = Polygon(box)
        distance = poly.area * unclip_ratio / poly.length
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded

    def get_mini_boxes(self, contour):

        if self.rotated_box:
            bounding_box = cv2.minAreaRect(contour)
            points = sorted(
                list(cv2.boxPoints(bounding_box)), key=lambda x: x[0]
            )

            index_1, index_2, index_3, index_4 = 0, 1, 2, 3
            if points[1][1] > points[0][1]:
                index_1 = 0
                index_4 = 1
            else:
                index_1 = 1
                index_4 = 0
            if points[3][1] > points[2][1]:
                index_2 = 2
                index_3 = 3
            else:
                index_2 = 3
                index_3 = 2

            box = [
                points[index_1],
                points[index_2],
                points[index_3],
                points[index_4],
            ]

            return box, min(bounding_box[1])

        else:
            l, t, w, h = cv2.boundingRect(contour)
            box = [[l, t], [l + w, t], [l + w, t + h], [l, t + h]]

            return box, min(w, h)

    def box_score_fast(self, bitmap, _box):
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype(np.int), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype(np.int), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype(np.int), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype(np.int), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype(np.int32), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def __call__(self, outs_dict, shape_list):

        textmap = outs_dict["text_map"]
        linkmap = outs_dict["link_map"]

        if torch.is_tensor(textmap):
            textmap = textmap.numpy()

        if torch.is_tensor(linkmap):
            linkmap = linkmap.numpy()

        text_threshold = self.text_thresh
        link_threshold = self.link_thresh

        use_dilate = self.use_dilate
        xdilate = self.xdilate
        ydilate = self.ydilate

        boxes_batch = []
        for batch_index in range(textmap.shape[0]):
            # perform thresholding on greyscale textmap
            _, text_score = cv2.threshold(
                textmap[batch_index], text_threshold, 1, 0
            )

            if use_dilate:
                # custom kernel defined to pad the textmap strictly in the upper left region of each blob
                center_x = int((xdilate - 1) / 2)
                center_y = int((ydilate - 1) / 2)

                inner = (
                    np.ones(center_x * center_y)
                    .reshape(center_y, center_x)
                    .astype(np.uint8)
                )
                outer_r = (
                    np.zeros((xdilate - center_x) * center_y)
                    .reshape(center_y, (xdilate - center_x))
                    .astype(np.uint8)
                )
                outer_d = (
                    np.zeros((xdilate) * -1 * (center_y - ydilate))
                    .reshape(ydilate - center_y, xdilate)
                    .astype(np.uint8)
                )

                final = np.append(outer_r, inner, 1)
                Vkernel = np.append(outer_d, final, 0)

                # dilation is performed here
                text_score = cv2.dilate(text_score, Vkernel, 1)

            # perform thresholding on greyscale linkmap
            _, link_score = cv2.threshold(
                linkmap[batch_index], link_threshold, 1, 0
            )
            text_score_comb = np.clip(text_score + link_score, 0, 1)

            src_h, src_w, _, _ = shape_list[batch_index]
            if self.merged_dilation_kernel is not None:
                mask = cv2.dilate(
                    np.array(text_score_comb).astype(np.uint8),
                    self.merged_dilation_kernel,
                )
            else:
                mask = text_score_comb

            boxes, _ = self.boxes_from_bitmap(
                textmap[batch_index], mask, src_w, src_h
            )

            boxes_batch.append({"points": boxes})

        return boxes_batch



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import sys

import cv2
import hydra
import numpy as np
import pandas as pd
from hydra.experimental import compose, initialize_config_dir
from omegaconf import OmegaConf
from PIL import Image
from tqdm import tqdm

__dir__ = os.path.dirname(__file__)
sys.path.append(os.path.join(__dir__, "../.."))

import ucr.inference.infer_cls as infer_cls
import ucr.inference.infer_det as infer_det
import ucr.inference.infer_rec as infer_rec
from ucr.utils.annotation import draw_ocr_box_txt, draw_text_det_res
from ucr.utils.utility import (
    check_and_read_gif,
    download_with_progressbar,
    get_image_file_list,
    merge_text_boxes,
    sorted_boxes,
)


class TextSystem(object):
    def __init__(self, config):
        self.text_detector = infer_det.TextDetector(config["Detection"])
        self.text_recognizer = infer_rec.TextRecognizer(config["Recognition"])
        self.text_classifier = infer_cls.TextClassifier(
            config["Classification"]
        )
        self.drop_score = config["drop_score"]
        self.merge_boxes = config["merge_boxes"]
        self.font_path = config["Recognition"]["font_path"]

        if self.merge_boxes:
            self.merge_slope_thresh = config["merge_slope_thresh"]
            self.merge_ycenter_thresh = config["merge_ycenter_thresh"]
            self.merge_height_thresh = config["merge_height_thresh"]
            self.merge_width_thresh = config["merge_width_thresh"]
            self.merge_add_margin = config["merge_add_margin"]

    def get_rotate_crop_image(self, img, points):
        """
        img_height, img_width = img.shape[0:2]
        left = int(np.min(points[:, 0]))
        right = int(np.max(points[:, 0]))
        top = int(np.min(points[:, 1]))
        bottom = int(np.max(points[:, 1]))
        img_crop = img[top:bottom, left:right, :].copy()
        points[:, 0] = points[:, 0] - left
        points[:, 1] = points[:, 1] - top
        """
        img_crop_width = int(
            max(
                np.linalg.norm(points[0] - points[1]),
                np.linalg.norm(points[2] - points[3]),
            )
        )
        img_crop_height = int(
            max(
                np.linalg.norm(points[0] - points[3]),
                np.linalg.norm(points[1] - points[2]),
            )
        )
        pts_std = np.float32(
            [
                [0, 0],
                [img_crop_width, 0],
                [img_crop_width, img_crop_height],
                [0, img_crop_height],
            ]
        )
        M = cv2.getPerspectiveTransform(points, pts_std)
        dst_img = cv2.warpPerspective(
            img,
            M,
            (img_crop_width, img_crop_height),
            borderMode=cv2.BORDER_REPLICATE,
            flags=cv2.INTER_CUBIC,
        )
        dst_img_height, dst_img_width = dst_img.shape[0:2]
        if dst_img_height * 1.0 / dst_img_width >= 1.5:
            dst_img = np.rot90(dst_img)
        return dst_img

    def perform_ocr(self, img, cls, verbose=False):
        ori_im = img.copy()
        dt_boxes, elapse = self.text_detector(img)
        if verbose:
            print(
                "\n------------------------dt_boxes num : {},\telapse : {:.3f}------------------------".format(
                    len(dt_boxes), elapse
                )
            )
        if dt_boxes is None:
            return None, None
        img_crop_list = []

        dt_boxes = sorted_boxes(dt_boxes)

        for bno in range(len(dt_boxes)):
            tmp_box = copy.deepcopy(dt_boxes[bno])
            img_crop = self.get_rotate_crop_image(ori_im, tmp_box)
            img_crop_list.append(img_crop)

        if cls:
            img_crop_list, _, elapse = self.text_classifier(img_crop_list)
            if verbose:
                print(
                    "------------------------cls num  : {},\telapse : {:.3f}------------------------".format(
                        len(img_crop_list), elapse
                    )
                )

        rec_res, elapse = self.text_recognizer(img_crop_list)
        if verbose:
            print(
                "------------------------rec_res num  : {},\telapse : {:.3f}------------------------\n".format(
                    len(rec_res), elapse
                )
            )

        filter_boxes, filter_rec_res = [], []

        if self.drop_score != 0.0:
            for box, rec_reuslt in zip(dt_boxes, rec_res):
                _, score = rec_reuslt
                if score >= self.drop_score:
                    filter_boxes.append(box)
                    filter_rec_res.append(rec_reuslt)
        else:
            filter_boxes = dt_boxes
            filter_rec_res = rec_res

        if self.merge_boxes:
            free_box, free_text, merged_box, merged_text = merge_text_boxes(
                filter_boxes,
                filter_rec_res,
                slope_thresh=self.merge_slope_thresh,
                ycenter_thresh=self.merge_ycenter_thresh,
                height_thresh=self.merge_height_thresh,
                width_thresh=self.merge_width_thresh,
                add_margin=self.merge_add_margin,
            )
            dt_boxes = free_box + merged_box
            rec_res = free_text + merged_text

        return dt_boxes, rec_res

    def __call__(
        self,
        input,
        output="output",
        o="output",
        det=True,
        rec=True,
        cls=False,
        return_type="df",
        save_image=False,
        save_tsv=True,
        verbose=False,
    ):
        assert isinstance(input, (np.ndarray, list, str))

        rec_dict = {}
        out_dict = {}

        if output != "output":
            output = output
        elif o != "output":
            output = o  # output takes higher precedence than o

        if save_tsv or save_image:
            print(
                "Saving prediction results in '{}' folder!\n".format(output),
                flush=True,
            )
            if not os.path.exists(output):
                os.makedirs(output)

        is_imgpath = False
        if isinstance(input, str):
            # download net image
            if input.startswith("http") and input.endswith(".jpg"):
                download_with_progressbar(input, "downloaded.jpg")
                input = "downloaded.jpg"

            input = get_image_file_list(input)
            is_imgpath = True

        elif isinstance(input, np.ndarray):
            input = [input]

        else:
            if isinstance(input[0], str):
                is_imgpath = True

        if len(input) == 0:
            print(
                "NO images found in {}. Please check input location!".format(
                    input
                )
            )
            sys.exit(0)

        print(f"Running Prediction on {len(input)} images:", flush=True)
        i = 0
        with tqdm(
            total=len(input),
            colour="green",
            desc="OCR",
            unit="image",
            unit_scale=True,
            position=0,
        ) as pbar:
            for image in input:
                if is_imgpath:
                    key = image
                    img, flag = check_and_read_gif(image)
                    if not flag:
                        img = cv2.imread(image)
                    if img is None:
                        print("ERROR in loading image:{}".format(image))
                        continue
                else:
                    key = str(i) + ".jpg"
                    i += 1
                    if len(image.shape) == 2:
                        img = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
                    elif len(image.shape) == 3:
                        img = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
                    elif len(image.shape) == 4:  # png file
                        img = cv2.cvtColor(image[:3], cv2.COLOR_RGB2BGR)
                    else:
                        print(
                            "ERROR: Input array has {} channels. Expected '2', '3' or '4'.".format(
                                len(image.shape)
                            )
                        )
                        continue
                if det:
                    if rec:
                        dt_boxes, rec_res = self.perform_ocr(
                            img, cls, verbose=verbose
                        )
                        if self.merge_boxes:
                            rec_res = rec_res[0:1]

                        if return_type == "list":
                            value = [
                                [box.tolist(), res]
                                for box, res in zip(dt_boxes, rec_res)
                            ]
                        elif return_type == "df" or save_tsv:
                            info_list = [
                                [
                                    int(box[0][0]),
                                    int(box[0][1]),
                                    int(box[2][0]),
                                    int(box[2][1]),
                                    res[0],
                                ]
                                for box, res in zip(dt_boxes, rec_res)
                            ]
                            value = pd.DataFrame(
                                info_list,
                                columns=[
                                    "startX",
                                    "startY",
                                    "endX",
                                    "endY",
                                    "Text",
                                ],
                            )
                            if save_tsv:
                                tsv_out = os.path.join(output, "tsv")
                                if not os.path.exists(tsv_out):
                                    os.makedirs(tsv_out)
                                tsv_path = os.path.join(
                                    tsv_out,
                                    "ocr_{}".format(
                                        os.path.splitext(
                                            os.path.basename(key)
                                        )[0]
                                    ),
                                )
                                value.to_csv(
                                    tsv_path + ".tsv", sep="\t", index=False
                                )

                        if verbose:
                            from tabulate import tabulate

                            headers = ["OCR Result", "Score"]
                            if self.merge_boxes:
                                print(
                                    tabulate(
                                        rec_res[0:1],
                                        headers,
                                        tablefmt="fancy_grid",
                                    )
                                )
                            else:
                                print(
                                    tabulate(
                                        rec_res, headers, tablefmt="fancy_grid"
                                    )
                                )

                        if save_image:
                            image = Image.fromarray(
                                cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                            )
                            boxes = dt_boxes
                            txts = [rec_res[i][0] for i in range(len(rec_res))]
                            scores = [
                                rec_res[i][1] for i in range(len(rec_res))
                            ]

                            draw_img = draw_ocr_box_txt(
                                image,
                                boxes,
                                txts,
                                scores,
                                drop_score=self.drop_score,
                                font_path=self.font_path,
                            )
                            img_out = os.path.join(output, "image")
                            if not os.path.exists(img_out):
                                os.makedirs(img_out)
                            img_path = os.path.join(
                                img_out,
                                "ocr_{}".format(
                                    os.path.splitext(os.path.basename(key))[0]
                                ),
                            )
                            cv2.imwrite(
                                img_path + ".jpg", draw_img[:, :, ::-1]
                            )
                            pbar.set_postfix_str(
                                "Output Path: '{}'".format(img_path)
                            )

                    else:
                        dt_boxes, elapse = self.text_detector(img)
                        if verbose:
                            print(
                                "\n------------------------dt_boxes num : {},\telapse : {:.3f}------------------------".format(
                                    len(dt_boxes), elapse
                                )
                            )
                        if dt_boxes is not None:
                            if return_type == "list":
                                value = [box.tolist() for box in dt_boxes]

                            elif return_type == "df" or save_tsv:
                                info_list = [
                                    [
                                        int(box[0][0]),
                                        int(box[0][1]),
                                        int(box[2][0]),
                                        int(box[2][1]),
                                    ]
                                    for box in dt_boxes
                                ]
                                value = pd.DataFrame(
                                    info_list,
                                    columns=[
                                        "startX",
                                        "startY",
                                        "endX",
                                        "endY",
                                    ],
                                )
                                if save_tsv:
                                    tsv_out = os.path.join(output, "tsv")
                                    if not os.path.exists(tsv_out):
                                        os.makedirs(tsv_out)
                                    tsv_path = os.path.join(
                                        tsv_out,
                                        "ocr_{}".format(
                                            os.path.splitext(
                                                os.path.basename(key)
                                            )[0]
                                        ),
                                    )
                                    value.to_csv(
                                        tsv_path + ".tsv",
                                        sep="\t",
                                        index=False,
                                    )

                            if save_image:
                                src_im = draw_text_det_res(dt_boxes, img)

                                img_out = os.path.join(output, "image")
                                if not os.path.exists(img_out):
                                    os.makedirs(img_out)
                                img_path = os.path.join(
                                    img_out,
                                    "det_{}".format(
                                        os.path.splitext(
                                            os.path.basename(key)
                                        )[0]
                                    ),
                                )
                                cv2.imwrite(img_path + ".jpg", src_im)
                                pbar.set_postfix_str(
                                    "Output Path: '{}'".format(img_path)
                                )

                    out_dict[key] = value
                else:
                    rec_dict[key] = img
                pbar.update(1)

        if rec_dict:
            img = list(rec_dict.values())
            if cls:
                img, cls_res, _ = self.text_classifier(img)
                if not rec:
                    if verbose:
                        cls_list = [
                            [k, v[0], v[1]]
                            for k, v in zip(list(rec_dict.keys()), cls_res)
                        ]
                        print("Classification Result:\n")
                        from tabulate import tabulate

                        headers = [
                            "File Name",
                            "Classification Result",
                            "Score",
                        ]
                        print(
                            tabulate(cls_list, headers, tablefmt="fancy_grid")
                        )

                    if save_image:
                        img_out = os.path.join(output, "image")
                        if not os.path.exists(img_out):
                            os.makedirs(img_out)
                        output_names = [
                            f"{k}_{v[0]}.jpg"
                            for k, v in zip(list(rec_dict.keys()), cls_res)
                        ]
                        for bno in range(len(img)):
                            cv2.imwrite(
                                os.path.join(img_out, output_names[bno]),
                                img[bno],
                            )

                    cls_output = {
                        k: [v] for k, v in zip(list(rec_dict.keys()), cls_res)
                    }
                    return cls_output
            if rec:
                rec_res, _ = self.text_recognizer(img)
                if verbose:
                    rec_list = [
                        [k, v[0], v[1]]
                        for k, v in zip(list(rec_dict.keys()), rec_res)
                    ]
                    print("Recognition Result:\n")
                    from tabulate import tabulate

                    headers = ["File Name", "Recognition Result", "Score"]
                    print(tabulate(rec_list, headers, tablefmt="fancy_grid"))

                if save_image:
                    img_out = os.path.join(output, "image")
                    if not os.path.exists(img_out):
                        os.makedirs(img_out)
                    output_names = [
                        f"{k}_{v[0]}.jpg"
                        for k, v in zip(list(rec_dict.keys()), rec_res)
                    ]
                    for bno in range(len(img)):
                        cv2.imwrite(
                            os.path.join(img_out, output_names[bno]), img[bno]
                        )

                rec_output = {
                    k: [v] for k, v in zip(list(rec_dict.keys()), rec_res)
                }
                return rec_output
        else:
            return out_dict


def main(args):
    cfg_dir = hydra.utils.to_absolute_path(args.config_dir)

    with initialize_config_dir(
        config_dir=cfg_dir, job_name=args.config_det_name
    ):
        cfg = compose(config_name=args.config_det_name)
        # print("Detection config:\n{}\n".format(cfg.pretty()))
        config_det = OmegaConf.to_container(cfg)

    with initialize_config_dir(
        config_dir=cfg_dir, job_name=args.config_rec_name
    ):
        cfg = compose(config_name=args.config_rec_name)
        # print("Recognition config:\n{}\n".format(cfg.pretty()))
        config_rec = OmegaConf.to_container(cfg)

    with initialize_config_dir(
        config_dir=cfg_dir, job_name=args.config_cls_name
    ):
        cfg = compose(config_name=args.config_cls_name)
        # print("Classification config:\n{}\n".format(cfg.pretty()))
        config_cls = OmegaConf.to_container(cfg)

    with initialize_config_dir(
        config_dir=cfg_dir, job_name=args.config_system_name
    ):
        cfg = compose(config_name=args.config_system_name)
        # print("Final config:\n{}\n".format(cfg.pretty()))
        config = OmegaConf.to_container(cfg)

    config["Detection"] = config_det
    config["Recognition"] = config_rec
    config["Classification"] = config_cls

    text_sys = TextSystem(config)

    input = hydra.utils.to_absolute_path(config["input"])
    _ = text_sys(
        input,
        output=hydra.utils.to_absolute_path(config["output"]),
        det=config["det"],
        rec=config["rec"],
        cls=config["cls"],
        return_type=config["return_type"],
        save_image=config["save_image"],
        save_tsv=config["save_tsv"],
        verbose=config["verbose"],
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--config_dir", type=str, default="ucr/conf")
    parser.add_argument("--config_det_name", type=str, default="infer_det")
    parser.add_argument("--config_rec_name", type=str, default="infer_rec")
    parser.add_argument("--config_cls_name", type=str, default="infer_cls")
    parser.add_argument(
        "--config_system_name", type=str, default="infer_system"
    )

    args = parser.parse_args()
    main(args)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
import os
import sys

import cv2
import numpy as np
from PIL import Image, ImageDraw, ImageFont

__dir__ = os.path.dirname(__file__)
sys.path.append(os.path.join(__dir__, "../.."))


def draw_text_det_res(dt_boxes, img):
    src_im = img.copy()
    for box in dt_boxes:
        box = np.array(box).astype(np.int32).reshape(-1, 2)
        cv2.polylines(src_im, [box], True, color=(20, 255, 57), thickness=2)
    return src_im


def resize_img(img, input_size=600):
    """
    resize img and limit the longest side of the image to input_size
    """
    img = np.array(img)
    im_shape = img.shape
    im_size_max = np.max(im_shape[0:2])
    im_scale = float(input_size) / float(im_size_max)
    img = cv2.resize(img, None, None, fx=im_scale, fy=im_scale)
    return img


def draw_ocr(
    image,
    boxes,
    txts=None,
    scores=None,
    drop_score=0.5,
    font_path="./ucr/utils/fonts/only_en.ttf",
):
    """
    Visualize the results of OCR detection and recognition
    args:
        image(Image|array): RGB image
        boxes(list): boxes with shape(N, 4, 2)
        txts(list): the texts
        scores(list): txxs corresponding scores
        drop_score(float): only scores greater than drop_threshold will be visualized
        font_path: the path of font which is used to draw text
    return(array):
        the visualized img
    """
    if scores is None:
        scores = [1] * len(boxes)
    box_num = len(boxes)
    for i in range(box_num):
        if scores is not None and (
            scores[i] < drop_score or math.isnan(scores[i])
        ):
            continue
        box = np.reshape(np.array(boxes[i]), [-1, 1, 2]).astype(np.int64)
        image = cv2.polylines(np.array(image), [box], True, (255, 0, 0), 2)
    if txts is not None:
        img = np.array(resize_img(image, input_size=600))
        txt_img = text_visual(
            txts,
            scores,
            img_h=img.shape[0],
            img_w=600,
            threshold=drop_score,
            font_path=font_path,
        )
        img = np.concatenate([np.array(img), np.array(txt_img)], axis=1)
        return img
    return image


def draw_ocr_box_txt(
    image,
    boxes,
    txts,
    scores=None,
    drop_score=0.5,
    font_path="./ucr/utils/fonts/only_en.ttf",
):
    h, w = image.height, image.width
    img_left = image.copy()
    img_right = Image.new("RGB", (w, h), (255, 255, 255))

    import random

    random.seed(0)
    draw_left = ImageDraw.Draw(img_left)
    draw_right = ImageDraw.Draw(img_right)
    for idx, (box, txt) in enumerate(zip(boxes, txts)):

        if scores is not None and scores[idx] < drop_score:
            continue
        color = (
            random.randint(0, 200),
            random.randint(0, 200),
            random.randint(0, 200),
        )
        draw_left.line(
            [
                (box[0][0], box[0][1]),
                (box[1][0] - 3, box[1][1]),
                (box[2][0] - 3, box[2][1]),
                (box[3][0], box[3][1]),
                (box[0][0], box[0][1]),
            ],
            fill=color,
            width=4,
        )

        box_height = math.sqrt(
            (box[0][0] - box[3][0]) ** 2 + (box[0][1] - box[3][1]) ** 2
        )
        box_width = math.sqrt(
            (box[0][0] - box[1][0]) ** 2 + (box[0][1] - box[1][1]) ** 2
        )
        if box_height > 2 * box_width:
            font_size = max(int(box_width * 0.9), 10)
            font = ImageFont.truetype(font_path, font_size, encoding="utf-8")
            cur_y = box[0][1]
            for c in txt:
                char_size = font.getsize(c)
                draw_right.text(
                    (box[0][0] + 3, cur_y), c, fill=(0, 0, 0), font=font
                )
                cur_y += char_size[1]
            draw_right.polygon(
                [
                    box[0][0],
                    box[0][1],
                    box[1][0] - 3,
                    box[1][1],
                    box[2][0] - 3,
                    box[2][1],
                    box[3][0],
                    box[3][1],
                ],
                outline=color,
            )
        else:
            img_fraction = 1
            # font_size=1
            font_change = False
            font_size = max(int(box_height * 0.8), 10)
            font = ImageFont.truetype(font_path, font_size)
            while font.getsize(txt)[0] > img_fraction * box_width:
                # iterate until the text size is just larger than the criteria
                font_change = True
                font_size -= 1
                font = ImageFont.truetype(font_path, font_size)
            if font_change:
                font_size += 1
            font = ImageFont.truetype(font_path, font_size)
            draw_right.text(
                [box[0][0] + 1, box[0][1] + 1], txt, fill=(0, 0, 0), font=font
            )

            wid, het = font.getsize(txt)
            draw_right.polygon(
                [
                    box[0][0],
                    box[0][1],
                    box[0][0] + wid,
                    box[0][1],
                    box[0][0] + wid,
                    box[0][1] + het + 2,
                    box[0][0],
                    box[0][1] + het + 2,
                ],
                outline=color,
            )
    img_left = Image.blend(image, img_left, 0.5)
    img_show = Image.new("RGB", (w * 2, h), (255, 255, 255))
    img_show.paste(img_left, (0, 0, w, h))
    img_show.paste(img_right, (w, 0, w * 2, h))
    return np.array(img_show)


def str_count(s):
    """
    Count the number of Chinese characters,
    a single English character and a single number
    equal to half the length of Chinese characters.
    args:
        s(string): the input of string
    return(int):
        the number of Chinese characters
    """
    import string

    count_zh = count_pu = 0
    s_len = len(s)
    en_dg_count = 0
    for c in s:
        if c in string.ascii_letters or c.isdigit() or c.isspace():
            en_dg_count += 1
        elif c.isalpha():
            count_zh += 1
        else:
            count_pu += 1
    return s_len - math.ceil(en_dg_count / 2)


def text_visual(
    texts,
    scores,
    img_h=400,
    img_w=600,
    threshold=0.0,
    font_path="./doc/simfang.ttf",
):
    """
    create new blank img and draw txt on it
    args:
        texts(list): the text will be draw
        scores(list|None): corresponding score of each txt
        img_h(int): the height of blank img
        img_w(int): the width of blank img
        font_path: the path of font which is used to draw text
    return(array):
    """
    if scores is not None:
        assert len(texts) == len(
            scores
        ), "The number of txts and corresponding scores must match"

    def create_blank_img():
        blank_img = np.ones(shape=[img_h, img_w], dtype=np.int8) * 255
        blank_img[:, img_w - 1 :] = 0
        blank_img = Image.fromarray(blank_img).convert("RGB")
        draw_txt = ImageDraw.Draw(blank_img)
        return blank_img, draw_txt

    blank_img, draw_txt = create_blank_img()

    font_size = 20
    txt_color = (0, 0, 0)
    font = ImageFont.truetype(font_path, font_size, encoding="utf-8")

    gap = font_size + 5
    txt_img_list = []
    count, index = 1, 0
    for idx, txt in enumerate(texts):
        index += 1
        if scores[idx] < threshold or math.isnan(scores[idx]):
            index -= 1
            continue
        first_line = True
        while str_count(txt) >= img_w // font_size - 4:
            tmp = txt
            txt = tmp[: img_w // font_size - 4]
            if first_line:
                new_txt = str(index) + ": " + txt
                first_line = False
            else:
                new_txt = "    " + txt
            draw_txt.text((0, gap * count), new_txt, txt_color, font=font)
            txt = tmp[img_w // font_size - 4 :]
            if count >= img_h // gap - 1:
                txt_img_list.append(np.array(blank_img))
                blank_img, draw_txt = create_blank_img()
                count = 0
            count += 1
        if first_line:
            new_txt = str(index) + ": " + txt + "   " + "%.3f" % (scores[idx])
        else:
            new_txt = "  " + txt + "  " + "%.3f" % (scores[idx])
        draw_txt.text((0, gap * count), new_txt, txt_color, font=font)
        # whether add new blank img or not
        if count >= img_h // gap - 1 and idx + 1 < len(texts):
            txt_img_list.append(np.array(blank_img))
            blank_img, draw_txt = create_blank_img()
            count = 0
        count += 1
    txt_img_list.append(np.array(blank_img))
    if len(txt_img_list) == 1:
        blank_img = np.array(txt_img_list[0])
    else:
        blank_img = np.concatenate(txt_img_list, axis=1)
    return np.array(blank_img)


def base64_to_cv2(b64str):
    import base64

    data = base64.b64decode(b64str.encode("utf8"))
    data = np.fromstring(data, np.uint8)
    data = cv2.imdecode(data, cv2.IMREAD_COLOR)
    return data


def draw_boxes(image, boxes, scores=None, drop_score=0.5):
    if scores is None:
        scores = [1] * len(boxes)
    for (box, score) in zip(boxes, scores):
        if score < drop_score:
            continue
        box = np.reshape(np.array(box), [-1, 1, 2]).astype(np.int64)
        image = cv2.polylines(np.array(image), [box], True, (255, 0, 0), 2)
    return image


def print_draw_crop_rec_res(output, filename, img_crop_list):
    bbox_num = len(img_crop_list)

    if not os.path.exists(output):
        os.makedirs(output)
    for bno in range(bbox_num):
        cv2.imwrite(
            os.path.join(output, f"{filename}_bno"), img_crop_list[bno]
        )


if __name__ == "__main__":
    test_img = "./doc/test_v2"
    predict_txt = "./doc/predict.txt"
    f = open(predict_txt, "r")
    data = f.readlines()
    img_path, anno = data[0].strip().split("\t")
    img_name = os.path.basename(img_path)
    img_path = os.path.join(test_img, img_name)
    image = Image.open(img_path)

    data = json.loads(anno)
    boxes, txts, scores = [], [], []
    for dic in data:
        boxes.append(dic["points"])
        txts.append(dic["transcription"])
        scores.append(round(dic["scores"], 3))

    new_img = draw_ocr(image, boxes, txts, scores)

    cv2.imwrite(img_name, new_img)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import logging
import os
import sys

import hydra
import numpy as np
from hydra.core.global_hydra import GlobalHydra
from omegaconf import OmegaConf

log = logging.getLogger(__name__)
__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, "../..")))

import time
import traceback

import cv2
import torch
from tqdm import tqdm

from ucr.core.architecture import build_architecture
from ucr.core.postprocess import build_postprocess
from ucr.core.preprocess import build_preprocess, preprocess
from ucr.core.preprocess.label_ops import BaseRecLabelEncode
from ucr.utils.utility import check_and_read_gif, get_image_file_list


class TextRecognizer(object):
    def __init__(self, config):
        self.device = config["device"]

        for op in config["Preprocess"]:
            op_name = list(op)[0]
            if op_name == "RecResizeImg":
                image_shape = op[op_name][
                    "image_shape"
                ]  # TODO:add try except here

        self.rec_image_shape = image_shape
        self.lang = config["lang"]
        self.rec_batch_num = config["batch_size"]
        self.rec_algorithm = config["Architecture"]["algorithm"]
        self.whitelist = config["whitelist"]
        self.blacklist = config["blacklist"]

        # Todo: Implement whitelist and blacklist in postprocessing step.

        char_dict_location = config["char_dict_location"]
        self.char_ops = BaseRecLabelEncode(
            config["max_text_length"],
            char_dict_location,
            config["lang"],
            config["use_space_char"],
        )

        global_keys = ["lang", "use_space_char", "max_text_length"]
        global_cfg = {
            key: value for key, value in config.items() if key in global_keys
        }
        global_cfg["char_dict_location"] = char_dict_location

        self.preprocess_op = build_preprocess(config["Preprocess"])
        self.postprocess_op = build_postprocess(
            config["Postprocess"], global_cfg
        )

        # build model
        if hasattr(self.postprocess_op, "character"):
            config["Architecture"]["Head"]["out_channels"] = len(
                getattr(self.postprocess_op, "character")
            )
        self.predictor = build_architecture(config["Architecture"])

        if self.device == "cuda":
            self.device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )
        elif self.device == "cpu":
            self.device = torch.device("cpu")
        else:
            log.error("wrong device selected! Choose eiter 'cuda' or 'cpu'")
            sys.exit(0)
        self.predictor.load_state_dict(
            torch.load(config["model_location"], map_location=self.device)
        )
        self.predictor.eval()

    def __call__(self, img_list):
        img_num = len(img_list)
        # Calculate the aspect ratio of all text bars
        width_list = []
        for img in img_list:
            width_list.append(img.shape[1] / float(img.shape[0]))
        # Sorting can speed up the recognition process
        indices = np.argsort(np.array(width_list))

        # rec_res = []
        rec_res = [["", 0.0]] * img_num
        batch_num = self.rec_batch_num
        elapse = 0
        for beg_img_no in range(0, img_num, batch_num):
            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            for ino in range(beg_img_no, end_img_no):

                image = {"image": img_list[indices[ino]]}
                norm_img = preprocess(image, self.preprocess_op)

                if self.rec_algorithm != "SRN":
                    norm_img = norm_img["image"][np.newaxis, :]
                    norm_img_batch.append(norm_img)
                else:
                    encoder_word_pos_list = []
                    gsrm_word_pos_list = []
                    gsrm_slf_attn_bias1_list = []
                    gsrm_slf_attn_bias2_list = []
                    encoder_word_pos_list.append(norm_img[1])
                    gsrm_word_pos_list.append(norm_img[2])
                    gsrm_slf_attn_bias1_list.append(norm_img[3])
                    gsrm_slf_attn_bias2_list.append(norm_img[4])
                    norm_img_batch.append(norm_img[0])
            norm_img_batch = np.concatenate(norm_img_batch)
            norm_img_batch = norm_img_batch.copy()

            if self.rec_algorithm == "SRN":
                starttime = time.time()
                encoder_word_pos_list = np.concatenate(encoder_word_pos_list)
                gsrm_word_pos_list = np.concatenate(gsrm_word_pos_list)
                gsrm_slf_attn_bias1_list = np.concatenate(
                    gsrm_slf_attn_bias1_list
                )
                gsrm_slf_attn_bias2_list = np.concatenate(
                    gsrm_slf_attn_bias2_list
                )
                inputs = [
                    torch.as_tensor(norm_img_batch, device=self.device),
                    torch.as_tensor(encoder_word_pos_list, device=self.device),
                    torch.as_tensor(gsrm_word_pos_list, device=self.device),
                    torch.as_tensor(
                        gsrm_slf_attn_bias1_list, device=self.device
                    ),
                    torch.as_tensor(
                        gsrm_slf_attn_bias2_list, device=self.device
                    ),
                ]
                self.predictor.to(self.device)
                with torch.no_grad():
                    output_tensors = self.predictor(inputs[0], inputs[1:])
                outputs = []
                for output_tensor in output_tensors.values():
                    output = output_tensor.cpu().data.numpy()
                    outputs.append(output)
                preds = {"predict": outputs[3]}
            else:
                starttime = time.time()
                input_tensors = torch.as_tensor(
                    norm_img_batch, device=self.device
                )
                self.predictor.to(self.device)
                with torch.no_grad():
                    output_tensors = self.predictor(input_tensors)

                preds = output_tensors.cpu().data.numpy()
                # If both blacklist and whitelist are provided, whitelist is only used
                if not self.whitelist and self.blacklist:
                    self.mod_chars = np.arange(preds.shape[-1])
                    black_list = self.char_ops.encode(self.blacklist)
                    black_list = np.array(black_list) + 1
                    self.mod_chars = np.setdiff1d(self.mod_chars, black_list)
                elif self.whitelist:
                    white_list = self.char_ops.encode(self.whitelist)
                    self.mod_chars = np.append(white_list, [-1]) + 1
                elif not self.whitelist and not self.blacklist:
                    self.mod_chars = []

                if len(self.mod_chars) != 0:
                    mod_onehot = np.zeros(preds.shape[-1])
                    mod_onehot[self.mod_chars] = 1
                    preds = np.multiply(
                        preds, mod_onehot
                    )  # * Implemented blacklist and whitelist here!

            rec_result = self.postprocess_op(preds)
            for rno in range(len(rec_result)):
                rec_res[indices[beg_img_no + rno]] = rec_result[rno]
            elapse += time.time() - starttime
        return rec_res, elapse


def main(cfg):
    log.debug("Recognition config:\n{}\n".format(cfg.pretty()))
    config = OmegaConf.to_container(cfg)
    GlobalHydra.instance().clear()

    input = hydra.utils.to_absolute_path(config["input"])
    model_location = hydra.utils.to_absolute_path(config["model_location"])
    config["model_location"] = model_location
    char_dict_location = hydra.utils.to_absolute_path(
        config["char_dict_location"]
    )
    config["char_dict_location"] = char_dict_location

    image_file_list = get_image_file_list(input)
    text_recognizer = TextRecognizer(config)
    valid_image_file_list = []
    img_list = []
    for image_file in tqdm(
        image_file_list, colour="green", desc="Recognition", unit="image"
    ):
        img, flag = check_and_read_gif(image_file)
        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            log.warning("error in loading image:{}".format(image_file))
            continue
        valid_image_file_list.append(image_file)
        img_list.append(img)
    try:
        rec_res, predict_time = text_recognizer(img_list)
    except Exception:
        log.error(traceback.format_exc())
        log.error(
            "ERROR!!!! \n"
            "Please read the FAQ：https://github.com/PaddlePaddle/PaddleOCR#faq \n"
            "If your model has tps module:  "
            "TPS does not support variable shape.\n"
            "Please set --rec_image_shape='3,32,100' and --rec_char_type='en' "
        )
        exit()
    for ino in range(len(img_list)):
        log.info(
            "[{}/{}] Prediction ['RESULT', SCORE] of {}:\t{}".format(
                ino + 1,
                len(img_list),
                valid_image_file_list[ino],
                rec_res[ino],
            )
        )
    log.info(
        "\nTotal Prediction time for {} images:\t{:.5f} s".format(
            len(img_list), predict_time
        )
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--config_path", type=str, default="../conf")
    parser.add_argument("--config_name", type=str, default="infer_rec")

    args = parser.parse_args()

    main_wrapper = hydra.main(
        config_path=args.config_path, config_name=args.config_name
    )
    main_wrapper(main)()



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import logging
import string

import numpy as np

log = logging.getLogger(__name__)


class ClsLabelEncode(object):
    def __init__(self, label_list, **kwargs):
        self.label_list = label_list

    def __call__(self, data):
        label = data["label"]
        if label not in self.label_list:
            return None
        label = self.label_list.index(label)
        data["label"] = label
        return data


class DetLabelEncode(object):
    def __init__(self, **kwargs):
        pass

    def __call__(self, data):
        import json

        label = data["label"]
        label = json.loads(label)
        nBox = len(label)
        boxes, txts, txt_tags = [], [], []
        for bno in range(0, nBox):
            box = label[bno]["points"]
            txt = label[bno]["transcription"]
            boxes.append(box)
            txts.append(txt)
            if txt in ["*", "###"]:
                txt_tags.append(True)
            else:
                txt_tags.append(False)
        boxes = self.expand_points_num(boxes)
        boxes = np.array(boxes, dtype=np.float32)
        txt_tags = np.array(txt_tags, dtype=np.bool)

        data["polys"] = boxes
        data["texts"] = txts
        data["ignore_tags"] = txt_tags
        return data

    def order_points_clockwise(self, pts):
        rect = np.zeros((4, 2), dtype="float32")
        s = pts.sum(axis=1)
        rect[0] = pts[np.argmin(s)]
        rect[2] = pts[np.argmax(s)]
        diff = np.diff(pts, axis=1)
        rect[1] = pts[np.argmin(diff)]
        rect[3] = pts[np.argmax(diff)]
        return rect

    def expand_points_num(self, boxes):
        max_points_num = 0
        for box in boxes:
            if len(box) > max_points_num:
                max_points_num = len(box)
        ex_boxes = []
        for box in boxes:
            ex_box = box + [box[-1]] * (max_points_num - len(box))
            ex_boxes.append(ex_box)
        return ex_boxes


class BaseRecLabelEncode(object):
    """ Convert between text-label and text-index """

    def __init__(
        self,
        max_text_length,
        char_dict_location=None,
        lang="ch_sim",
        use_space_char=False,
    ):
        support_lang = [
            "ch_sim",
            "en",
            "en_case",
            "en_number",
            "fr",
            "de",
            "ja",
            "ko",
            "it",
            "es",
            "pt",
            "ru",
            "ar",
            "hi",
            "ug",
            "fa",
            "ur",
            "rs_latin",
            "oc",
            "rs_cyrillic",
            "bg",
            "uk",
            "be",
            "te",
            "kn",
            "ch_tra",
            "ta",
            "mr",
            "ne",
        ]
        assert (
            lang in support_lang
        ), "Only {} are supported now but get {}".format(support_lang, lang)

        self.max_text_len = max_text_length
        self.beg_str = "sos"
        self.end_str = "eos"
        if lang == "en_case":
            self.character_str = "0123456789abcdefghijklmnopqrstuvwxyz"
            dict_character = list(self.character_str)
        elif lang == "en_number":
            # same with ASTER setting (use 94 char).
            self.character_str = string.printable[:62]
            if use_space_char:
                self.character_str += " "
            dict_character = list(self.character_str)
        elif lang in support_lang:
            self.character_str = ""
            assert (
                char_dict_location is not None
            ), "char_dict_location should not be None when lang is {}".format(
                lang
            )
            with open(char_dict_location, "rb") as fin:
                lines = fin.readlines()
                for line in lines:
                    line = line.decode("utf-8").strip("\n").strip("\r\n")
                    self.character_str += line
            if use_space_char:
                self.character_str += " "
            dict_character = list(self.character_str)
        self.lang = lang
        dict_character = self.add_special_char(dict_character)
        self.dict = {}
        for i, char in enumerate(dict_character):
            self.dict[char] = i
        self.character = dict_character

    def add_special_char(self, dict_character):
        return dict_character

    def encode(self, text):
        """convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]

        output:
            text: concatenated text index for CTCLoss.
                    [sum(text_lengths)] = [text_index_0 + text_index_1 + ... + text_index_(n - 1)]
            length: length of each text. [batch_size]
        """
        if len(text) == 0:
            return None
        if self.lang == "en_case":
            text = text.lower()
        text_list = []
        for char in text:
            if char not in self.dict:
                log.warning("{} is not in dict".format(char))
                continue
            text_list.append(self.dict[char])
        if len(text_list) == 0:
            return None
        return text_list


class CTCLabelEncode(BaseRecLabelEncode):
    """ Convert between text-label and text-index """

    def __init__(
        self,
        max_text_length,
        char_dict_location=None,
        lang="ch",
        use_space_char=False,
        **kwargs
    ):
        super(CTCLabelEncode, self).__init__(
            max_text_length, char_dict_location, lang, use_space_char
        )

    def __call__(self, data):
        text = data["label"]
        text = self.encode(text)
        if text is None:
            return None
        data["length"] = np.array(len(text))
        text = text + [0] * (self.max_text_len - len(text))
        data["label"] = np.array(text)
        return data

    def add_special_char(self, dict_character):
        dict_character = ["blank"] + dict_character
        return dict_character


class AttnLabelEncode(BaseRecLabelEncode):
    """ Convert between text-label and text-index """

    def __init__(
        self,
        max_text_length,
        char_dict_location=None,
        lang="ch",
        use_space_char=False,
        **kwargs
    ):
        super(AttnLabelEncode, self).__init__(
            max_text_length, char_dict_location, lang, use_space_char
        )

    def add_special_char(self, dict_character):
        self.beg_str = "sos"
        self.end_str = "eos"
        dict_character = [self.beg_str] + dict_character + [self.end_str]
        return dict_character

    def __call__(self, data):
        text = data["label"]
        text = self.encode(text)
        if text is None:
            return None
        if len(text) >= self.max_text_len:
            return None
        data["length"] = np.array(len(text))
        text = (
            [0]
            + text
            + [len(self.character) - 1]
            + [0] * (self.max_text_len - len(text) - 2)
        )
        data["label"] = np.array(text)
        return data

    def get_ignored_tokens(self):
        beg_idx = self.get_beg_end_flag_idx("beg")
        end_idx = self.get_beg_end_flag_idx("end")
        return [beg_idx, end_idx]

    def get_beg_end_flag_idx(self, beg_or_end):
        if beg_or_end == "beg":
            idx = np.array(self.dict[self.beg_str])
        elif beg_or_end == "end":
            idx = np.array(self.dict[self.end_str])
        else:
            assert False, (
                "Unsupport type %s in get_beg_end_flag_idx" % beg_or_end
            )
        return idx


class SRNLabelEncode(BaseRecLabelEncode):
    """ Convert between text-label and text-index """

    def __init__(
        self,
        max_text_length=25,
        char_dict_location=None,
        lang="en_case",
        use_space_char=False,
        **kwargs
    ):
        super(SRNLabelEncode, self).__init__(
            max_text_length, char_dict_location, lang, use_space_char
        )

    def add_special_char(self, dict_character):
        dict_character = dict_character + [self.beg_str, self.end_str]
        return dict_character

    def __call__(self, data):
        text = data["label"]
        text = self.encode(text)
        char_num = len(self.character)
        if text is None:
            return None
        if len(text) > self.max_text_len:
            return None
        data["length"] = np.array(len(text))
        text = text + [char_num - 1] * (self.max_text_len - len(text))
        data["label"] = np.array(text)
        return data

    def get_ignored_tokens(self):
        beg_idx = self.get_beg_end_flag_idx("beg")
        end_idx = self.get_beg_end_flag_idx("end")
        return [beg_idx, end_idx]

    def get_beg_end_flag_idx(self, beg_or_end):
        if beg_or_end == "beg":
            idx = np.array(self.dict[self.beg_str])
        elif beg_or_end == "end":
            idx = np.array(self.dict[self.end_str])
        else:
            assert False, (
                "Unsupport type %s in get_beg_end_flag_idx" % beg_or_end
            )
        return idx



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import sys

import torch
from hydra.experimental import compose, initialize_config_dir
from omegaconf import OmegaConf
from tabulate import tabulate

from ucr.inference import infer_system
from ucr.utils.utility import download_font, maybe_download

__all__ = ["UCR"]

__dir__ = os.path.dirname(os.path.abspath(__file__))
SUPPORT_DET_MODEL = ["DB", "CRAFT"]
SUPPORT_REC_MODEL = ["CRNN"]
SUPPORT_CLS_MODEL = ["CLS"]
VERSION = 1.0
SUPPORT_MODEL_TYPE = [
    "torch_mobile",
    "torch_server",
    "onnx_mobile",
    "onnx_server",
]
BASE_DIR = os.path.expanduser("~/.ucr/")


def print_version() -> None:
    """Prints version information of rasa tooling and python."""

    print(
        "UCR Version      :          0.2.14"
    )  # TODO: Change this to update automatically
    print(f"Python Version    :         {platform.python_version()}")
    print(f"Operating System  :         {platform.platform()}")
    print(f"Python Path       :         {sys.executable}")


def parse_args(mMain=True, add_help=True):
    import argparse

    def str2bool(v):
        return v.lower() in ("true", "t", "1")

    if mMain:
        parser = argparse.ArgumentParser(
            prog="ucr",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            description="UCR is an Open Source, Easy to use Python library to build Production Ready OCR applications with its highly Intuitive,\
            Modular & Extensible API design and off-the-shelf Pretrained Models for over 25 languages. For more details, please go to https://ucr.docyard.ai",
        )

        parser.add_argument(
            "-v",
            "--version",
            action="store_true",
            default=argparse.SUPPRESS,
            help="Print installed UCR version",
        )

        parent_parser = argparse.ArgumentParser(add_help=False)
        parent_parsers = [parent_parser]

        subparsers = parser.add_subparsers(help="UCR commands")

        predict_parser = subparsers.add_parser(
            "predict",
            parents=parent_parsers,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help="Loads pretrained models for performing predictions!",
        )
        test_parser = subparsers.add_parser(
            "test",
            parents=parent_parsers,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            help="Check if everything is installed and working correctly!",
        )
        # params for prediction system
        predict_parser.add_argument(
            "input", type=str, help="Input file/folder path"
        )
        predict_parser.add_argument("--conf_location", type=str, default=None)
        predict_parser.add_argument(
            "--force_download", type=str2bool, default=False
        )
        predict_parser.add_argument(
            "-o", "--output", type=str, default="output"
        )
        predict_parser.add_argument("-d", "--device", type=str, default="cuda")
        predict_parser.add_argument("-l", "--lang", type=str, default="ch_sim")
        predict_parser.add_argument("--save_tsv", type=str2bool, default=True)
        predict_parser.add_argument(
            "--save_image", type=str2bool, default=False
        )
        predict_parser.add_argument("--backend", type=str, default="torch")
        predict_parser.add_argument("--type", type=str, default="mobile")
        predict_parser.add_argument("--return_type", type=str, default="df")
        predict_parser.add_argument(
            "--system_overrides", nargs="+", type=str, default=[]
        )
        predict_parser.add_argument("--det", type=str2bool, default=True)
        predict_parser.add_argument("--rec", type=str2bool, default=True)
        predict_parser.add_argument("--cls", type=str2bool, default=False)
        predict_parser.add_argument("--verbose", type=str2bool, default=False)

        # params for detection engine
        predict_parser.add_argument(
            "--det_algorithm", type=str, default="CRAFT"
        )
        predict_parser.add_argument(
            "--det_config_name", type=str, default="infer_det"
        )
        predict_parser.add_argument(
            "--det_model_location", type=str, default=None
        )
        predict_parser.add_argument(
            "--det_batch_size", type=int, default=1
        )  # batch_size is not yet implemented in the code
        predict_parser.add_argument(
            "--det_overrides", nargs="+", type=str, default=[]
        )

        # params for recogniton engine
        predict_parser.add_argument(
            "--rec_algorithm", type=str, default="CRNN"
        )
        predict_parser.add_argument(
            "--rec_config_name", type=str, default="infer_rec"
        )
        predict_parser.add_argument(
            "--rec_model_location", type=str, default=None
        )
        predict_parser.add_argument("--rec_batch_size", type=int, default=8)
        predict_parser.add_argument(
            "--rec_overrides", nargs="+", type=str, default=[]
        )
        predict_parser.add_argument("--whitelist", type=str, default=None)
        predict_parser.add_argument("--blacklist", type=str, default=None)

        # params for detection engine
        predict_parser.add_argument("--cls_algorithm", type=str, default="CLS")
        predict_parser.add_argument(
            "--cls_config_name", type=str, default="infer_cls"
        )
        predict_parser.add_argument(
            "--cls_model_location", type=str, default=None
        )
        predict_parser.add_argument("--cls_batch_size", type=int, default=8)
        predict_parser.add_argument(
            "--cls_overrides", nargs="+", type=str, default=[]
        )
        predict_parser.set_defaults(func=predict_cli)

        test_parser.add_argument("-l", "--lang", type=str, default="ch_sim")
        test_parser.set_defaults(func=test_cli)
        return parser
    else:
        return argparse.Namespace(
            conf_location=None,
            force_download=False,
            d="cuda",
            device="cuda",
            l="ch_sim",
            lang="ch_sim",
            type="mobile",
            backend="torch",
            system_overrides=[],
            det_algorithm="CRAFT",
            det_config_name="infer_det",
            det_model_location=None,
            det_batch_size=1,
            det_overrides=[],
            rec_algorithm="CRNN",
            rec_config_name="infer_rec",
            rec_model_location=None,
            rec_batch_size=8,
            rec_overrides=[],
            whitelist=None,
            blacklist=None,
            cls_algorithm="CLS",
            cls_config_name="infer_cls",
            cls_model_location=None,
            cls_batch_size=8,
            cls_overrides=[],
        )


class UCR(infer_system.TextSystem):
    def __init__(self, **kwargs):
        """
        ucr package
        args:
            **kwargs: other params show in ucr --help
        """
        args = parse_args(mMain=False, add_help=False)
        args.__dict__.update(**kwargs)

        force_download = args.force_download
        conf_location = args.conf_location
        det_algorithm = args.det_algorithm
        args.rec_algorithm
        args.cls_algorithm

        type = args.type
        backend = args.backend
        model_type = backend + "_" + type

        lang = args.l
        if args.lang != "ch_sim":
            lang = args.lang
            if args.l != "ch_sim":
                print(
                    f'WARNING: "l (={args.l})" argument is superseded by "lang (={args.lang})" argument! l=lang={lang}'
                )

        device = args.d
        if args.device != "cuda":
            device = args.device
            if args.d != "cuda":
                print(
                    f'WARNING: "d (={args.d})" argument is superseded by "device (={args.device})" argument! d=device={device}'
                )

        if device == "cuda" and not torch.cuda.is_available():
            print(
                "WARNING: Unable to load CUDA kernels. Check if NVIDIA GPU is available and CUDA installed.\n \
                  Falling back to CPU execution."
            )
            device = "cpu"

        __dir__ = os.path.dirname(__file__)
        if not conf_location:
            conf_location = os.path.join(__dir__, "conf")

        with initialize_config_dir(
            config_dir=conf_location, job_name="infer_det"
        ):
            model_type = (
                "torch_server" if det_algorithm == "CRAFT" else model_type
            )
            overrides = [
                f"{k}={v}"
                for k, v in model_urls[model_type]["det"][
                    det_algorithm
                ].items()
                if k != "url"
            ]
            overrides.extend(args.det_overrides)
            cfg = compose(
                config_name=args.det_config_name, overrides=overrides
            )
            # print("Detection config:\n{}\n".format(cfg.pretty()))

            config_det = OmegaConf.to_container(cfg)
            model_type = backend + "_" + type

        with initialize_config_dir(
            config_dir=conf_location, job_name="infer_rec"
        ):
            model_type = "torch_mobile" if lang != "ch_sim" else model_type
            overrides = [
                f"{k}={v}"
                for k, v in model_urls[model_type]["rec"][lang].items()
                if k != "url"
            ]
            overrides.extend(args.rec_overrides)
            cfg = compose(
                config_name=args.rec_config_name, overrides=overrides
            )
            # print("Recognition config:\n{}\n".format(cfg.pretty()))

            config_rec = OmegaConf.to_container(cfg)
            model_type = backend + "_" + type

        with initialize_config_dir(
            config_dir=conf_location, job_name="infer_cls"
        ):
            overrides = [
                f"{k}={v}"
                for k, v in model_urls["torch_mobile"]["cls"].items()
                if k != "url"
            ]
            overrides.extend(args.cls_overrides)
            cfg = compose(
                config_name=args.cls_config_name, overrides=overrides
            )
            # print("Classification config:\n{}\n".format(cfg.pretty()))

            config_cls = OmegaConf.to_container(cfg)

        with initialize_config_dir(
            config_dir=conf_location, job_name="infer_system"
        ):
            cfg = compose(
                config_name="infer_system", overrides=args.system_overrides
            )
            # print("End-to-end config:\n{}\n".format(cfg.pretty()))

            config = OmegaConf.to_container(cfg)

        assert (
            model_type in SUPPORT_MODEL_TYPE
        ), "TYPE_BACKEND must be of {} format, but got {}".format(
            SUPPORT_MODEL_TYPE, model_type
        )

        config_det["device"] = config_rec["device"] = config_cls[
            "device"
        ] = device

        config_det["batch_size"] = args.det_batch_size
        config_rec["batch_size"] = args.rec_batch_size
        config_cls["batch_size"] = args.cls_batch_size

        config_rec["whitelist"] = args.whitelist
        config_rec["blacklist"] = args.blacklist
        config_rec["lang"] = lang

        # init model dir
        if not config_det["model_location"]:
            model_type = (
                "torch_server" if det_algorithm == "CRAFT" else model_type
            )
            det_path = maybe_download(
                os.path.join(BASE_DIR, "{}/det".format(VERSION)),
                model_urls[model_type]["det"][det_algorithm]["url"],
                force_download,
            )
            config_det["model_location"] = os.path.join(det_path, "model.pt")
            model_type = backend + "_" + type
        else:
            config_det["model_location"] = config_det[
                "model_location"
            ]  # add abs path in parser args.

        if not config_rec["model_location"]:
            model_type = "torch_mobile" if lang != "ch_sim" else model_type
            rec_path = maybe_download(
                os.path.join(BASE_DIR, "{}/rec/{}".format(VERSION, lang)),
                model_urls[model_type]["rec"][lang]["url"],
                force_download,
            )
            config_rec["model_location"] = os.path.join(rec_path, "model.pt")
            config_rec["font_path"] = rec_path = download_font(
                BASE_DIR,
                model_urls[model_type]["rec"][lang]["font_path"],
                force_download,
            )
            config_rec["char_dict_location"] = os.path.join(
                __dir__, config_rec["char_dict_location"]
            )
        else:
            config_rec["model_location"] = config_rec[
                "model_location"
            ]  # add abs path in parser args.
            config_rec["font_path"] = config_rec[
                "font_path"
            ]  # add abs path in .yaml file
            config_rec["char_dict_location"] = config_rec[
                "char_dict_location"
            ]  # add abs path in .yaml file

        if not config_cls["model_location"]:
            cls_path = maybe_download(
                os.path.join(BASE_DIR, "{}/cls".format(VERSION)),
                model_urls["torch_mobile"]["cls"]["url"],
                force_download,
            )
            config_cls["model_location"] = os.path.join(cls_path, "model.pt")
        else:
            config_cls["model_location"] = config_cls[
                "model_location"
            ]  # add abs path in parser args.

        config["Detection"] = config_det
        config["Recognition"] = config_rec
        config["Classification"] = config_cls

        super().__init__(config)

    def predict(
        self,
        input,
        output="output",
        o="output",
        det=True,
        rec=True,
        cls=False,
        return_type="df",
        save_image=False,
        save_tsv=True,
        verbose=False,
    ):
        return super().__call__(
            input,
            output,
            o,
            det,
            rec,
            cls,
            return_type,
            save_image,
            save_tsv,
            verbose,
        )


def test_cli(args):
    import numpy as np

    try:
        print(
            "\n------------------------Loading Pretrained Models------------------------"
        )
        ocr_engine = UCR(**(args.__dict__))
        print(
            "------------------------Loading SUCCESSFUL------------------------\n"
        )
        print(
            "\n------------------------[0/3] Testing Detection Module------------------------"
        )
        _ = ocr_engine.predict(
            np.random.randint(
                low=0, high=254, size=(10, 10, 3), dtype=np.uint8
            ),
            det=True,
            rec=False,
            cls=False,
            return_type="df",
            save_image=False,
            save_tsv=False,
            verbose=False,
        )
        print(
            "------------------------Detection SUCCESSFUL------------------------\n"
        )

        print(
            "\n------------------------[2/3] Testing Angle Correction Module------------------------"
        )
        _ = ocr_engine.predict(
            np.random.randint(
                low=0, high=254, size=(10, 10, 3), dtype=np.uint8
            ),
            det=False,
            rec=False,
            cls=True,
            return_type="df",
            save_image=False,
            save_tsv=False,
            verbose=False,
        )
        print(
            "------------------------Angle Correction SUCCESSFUL------------------------\n"
        )
        print(
            "\n------------------------[3/3] Testing Recognition Module------------------------"
        )
        _ = ocr_engine.predict(
            np.random.randint(
                low=0, high=254, size=(10, 10, 3), dtype=np.uint8
            ),
            det=False,
            rec=True,
            cls=False,
            return_type="df",
            save_image=False,
            save_tsv=False,
            verbose=False,
        )
        print(
            "------------------------Recognition SUCCESSFUL------------------------\n"
        )
        print(
            "\n:::::::::::::::::::::::::::::::::::::::: All tests passed! SUCCESS! ::::::::::::::::::::::::::::::::::::::::\n"
        )
    except Exception as e:
        print(
            f"\n:::::::::::::::::::::::::::::::::::::::: FAILED :::::::::::::::::::::::::::::::::::::::: \n{e}\n"
        )


def predict_cli(args):
    ocr_engine = UCR(**(args.__dict__))
    result = ocr_engine.predict(
        args.input,
        output=args.output,
        det=args.det,
        rec=args.rec,
        cls=args.cls,
        return_type=args.return_type,
        save_image=args.save_image,
        save_tsv=args.save_tsv,
        verbose=args.verbose,
    )
    if result is not None:
        for k, v in result.items():
            print(f"\n------------------------{k}:------------------------")
            print(tabulate(v, tablefmt="fancy_grid"))


def main():
    arg_parser = parse_args(mMain=True)
    cmdline_arguments = arg_parser.parse_args()

    try:
        if hasattr(cmdline_arguments, "input"):
            cmdline_arguments.func(cmdline_arguments)
        elif hasattr(cmdline_arguments, "lang"):
            cmdline_arguments.func(cmdline_arguments)
        elif hasattr(cmdline_arguments, "version"):
            print_version()
        else:
            # user has not provided a subcommand, let's print the help
            print("No command specified!")
            arg_parser.print_help()
            sys.exit(1)
    except Exception as e:
        print(f"{e.__class__.__name__}: {e}")
        sys.exit(1)


model_urls = {
    "torch_server": {
        "det": {
            "DB": {
                "preprocess": "det_db",
                "architecture": "det_ppocr_server",
                "postprocess": "det_db",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/det_ench_ppocr_server.zip",
            },
            "CRAFT": {
                "preprocess": "det_craft",
                "architecture": "det_vgg_craft",
                "postprocess": "det_craft",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/det_vgg_craft.zip",
            },
        },
        "rec": {
            "ch_sim": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_server",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/noto_cjk.otf",
                "char_dict_location": "utils/dict/ch_sim_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_server/rec_ench_ppocr_server.zip",
            }
        },
    },
    "torch_mobile": {
        "det": {
            "DB": {
                "preprocess": "det_db",
                "architecture": "det_ppocr_mobile",
                "postprocess": "det_db",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/det_ench_ppocr_mobile.zip",
            }
        },
        "rec": {
            "ch_sim": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/noto_cjk.otf",
                "char_dict_location": "utils/dict/ch_sim_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ench_ppocr_mobile.zip",
            },
            "en_number": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/only_en.ttf",
                "char_dict_location": "",  # won't be reqd/used as en_number dict would be constructed in code.
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_en_number_mobile.zip",
            },
            "fr": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/french.ttf",
                "char_dict_location": "utils/dict/french_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_french_mobile.zip",
            },
            "de": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/german.ttf",
                "char_dict_location": "utils/dict/german_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_german_mobile.zip",
            },
            "ko": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/noto_cjk.otf",
                "char_dict_location": "utils/dict/korean_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_korean_mobile.zip",
            },
            "ja": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/noto_cjk.otf",
                "char_dict_location": "utils/dict/japan_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_japan_mobile.zip",
            },
            "ch_tra": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/noto_cjk.otf",
                "char_dict_location": "utils/dict/chinese_cht_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_chinese_cht_mobile.zip",
            },
            "it": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/italian.ttf",
                "char_dict_location": "utils/dict/it_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_it_mobile.zip",
            },
            "es": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/spanish.ttf",
                "char_dict_location": "utils/dict/spanish_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_xi_mobile.zip",
            },
            "pt": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/portugese.ttf",
                "char_dict_location": "utils/dict/pu_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_pu_mobile.zip",
            },
            "ru": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/russian.ttf",
                "char_dict_location": "utils/dict/ru_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ru_mobile.zip",
            },
            "ar": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/arabic.ttf",
                "char_dict_location": "utils/dict/ar_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ar_mobile.zip",
            },
            "hi": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/hindi.ttf",
                "char_dict_location": "utils/dict/hi_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_hi_mobile.zip",
            },
            "ug": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/uyghur.ttf",
                "char_dict_location": "utils/dict/ug_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ug_mobile.zip",
            },
            "fa": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/persian.ttf",
                "char_dict_location": "utils/dict/fa_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_fa_mobile.zip",
            },
            "ur": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/urdu.ttf",
                "char_dict_location": "utils/dict/ur_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ur_mobile.zip",
            },
            "rs_latin": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/rs_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_rs_mobile.zip",
            },
            "oc": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/oc_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_oc_mobile.zip",
            },
            "mr": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/marathi.ttf",
                "char_dict_location": "utils/dict/mr_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_mr_mobile.zip",
            },
            "ne": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/nepali.ttf",
                "char_dict_location": "utils/dict/ne_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ne_mobile.zip",
            },
            "rs_cyrillic": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/rsc_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_rsc_mobile.zip",
            },
            "bg": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/bg_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_bg_mobile.zip",
            },
            "uk": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/uk_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_uk_mobile.zip",
            },
            "be": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/serbian.ttf",
                "char_dict_location": "utils/dict/be_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_be_mobile.zip",
            },
            "te": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/telugu.ttf",
                "char_dict_location": "utils/dict/te_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_te_mobile.zip",
            },
            "kn": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/kannada.ttf",
                "char_dict_location": "utils/dict/ka_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ka_mobile.zip",
            },
            "ta": {
                "preprocess": "rec_ctc",
                "architecture": "rec_ppocr_mobile",
                "postprocess": "rec_ctc",
                "font_path": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/fonts/tamil.ttf",
                "char_dict_location": "utils/dict/ta_dict.txt",
                "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/rec_ta_mobile.zip",
            },
        },
        "cls": {
            "preprocess": "cls",
            "architecture": "cls_ppocr_mobile",
            "postprocess": "cls",
            "url": "https://docyard.s3.us-west-000.backblazeb2.com/UCR/torch_mobile/cls_ench_ppocr_mobile.zip",
        },
    },
}



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["build_head"]


def build_head(config):
    # det head
    # cls head
    from .cls_head import ClsHead
    from .det_craft_head import CRAFTHead
    from .det_db_head import DBHead
    from .det_east_head import EASTHead
    from .det_sast_head import SASTHead
    from .rec_att_head import AttentionHead

    # rec head
    from .rec_ctc_head import CTCHead
    from .rec_srn_head import SRNHead

    support_dict = [
        "CRAFTHead",
        "DBHead",
        "EASTHead",
        "SASTHead",
        "CTCHead",
        "ClsHead",
        "AttentionHead",
        "SRNHead",
    ]

    module_name = config.pop("name")
    assert module_name in support_dict, Exception(
        "head only support {}".format(support_dict)
    )
    module_class = eval(module_name)(**config)
    return module_class



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ucr',
 'ucr.core',
 'ucr.core.architecture',
 'ucr.core.architecture.backbone',
 'ucr.core.architecture.head',
 'ucr.core.architecture.neck',
 'ucr.core.architecture.transform',
 'ucr.core.dataloader',
 'ucr.core.postprocess',
 'ucr.core.preprocess',
 'ucr.core.preprocess.text_image_aug',
 'ucr.inference',
 'ucr.utils']

package_data = \
{'': ['*'],
 'ucr': ['conf/*',
         'conf/architecture/*',
         'conf/hydra/job_logging/*',
         'conf/hydra/output/*',
         'conf/postprocess/*',
         'conf/preprocess/*'],
 'ucr.utils': ['dict/*']}

install_requires = \
['PyMuPDF>=1.18.10,<2.0.0',
 'hydra-core>=1.0.6,<2.0.0',
 'imgaug>=0.4.0,<0.5.0',
 'numpy<1.20.0',
 'opencv-python>=4.1.0,<5.0.0',
 'pandas>=1.1,<1.2',
 'pyclipper>=1.2.1,<2.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tqdm>=4.40,<5.0']

entry_points = \
{'console_scripts': ['ucr = ucr.ucr:main']}

setup_kwargs = {
    'name': 'ucr',
    'version': '0.2.15',
    'description': 'Universal Character Recognizer (UCR): Simple, Intuitive, Extensible, Multi-Lingual OCR engine',
    'long_description': '<br> <br>\n<p align="center"><img src="docs/static/images/VectorU.svg" alt="Github Runner Covergae Status" height="100">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<img src="docs/static/images/VectorC.svg" alt="Github Runner Covergae Status" height="100">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<img src="docs/static/images/VectorR.svg" alt="Github Runner Covergae Status" height="100"></p>\n<br> <br>\n<p align="center">Universal Character Recognizer (UCR) is an <u>Open Source</u>, <u>Easy to use</u> Python library to build <u>Production Ready</u> OCR applications with its highly Intuitive,  Modular & Extensible API design and off-the-shelf <a href="docs/modelzoo.md">Pretrained Models</a> for over <b>25 languages</b>.</p>\n<p align="center">\n  Read UCR Documentation on <u><a href="https://ucr.docyard.ai/">ucr.docyard.ai</a></u>\n  <br> <br>\n  <a href="#about">Features</a> •\n  <a href="#setup">Setup</a> •\n  <a href="#usage">Usage</a> •\n  <a href="#acknowledgement">Acknowledgement</a>\n  <br> <br>\n  <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/ucr">\n  <a href="https://badge.fury.io/py/ucr"><img src="https://badge.fury.io/py/ucr.svg" alt="PyPI version"></a>\n</p>\n\n## Demo\n\n#### For details, click [here](https://ucr.docyard.ai/demo)!\n\n<p align="center"><img src="docs/static/images/demo.gif"/></p>\n\n## Features\n\n- Supports SOTA Text Detection and Recognition models\n- Built on top of Pytorch and Pytorch Lightning\n- Supports over 25 languages\n- Model Zoo contains 27 Pretrained Models across 25 languages\n- Modular Design Language allows Pick and Choose of different components\n- Easily extensible with Custom Components and attributes\n- Hydra config enables Rapid Prototyping with multiple configurations\n- Support for Packaging, Logging and Deployment tools straight out of the box\n\n*Note: Some features are still in active development and might not be available.*\n## Setup\n\n### Installation\n\n**Require python version >= 3.6.2, install with `pip` (recommended)**\n\n1. <b>Prerequisites:</b> Install compatible version of Pytorch and torchvision from [official repository](https://pytorch.org/get-started/locally/).\n2. <b>Installation:</b> Install the latest stable version of UCR:\n```shell\npip install -U ucr\n```\n\n#### <span style="color:#FF8856">[Optional]</span> Test Installation\n\nRun dummy tests!\n```python\nucr test\n# Optional: Add -l/--lang=\'language_id\' to test on particular language!\nucr test -l=\'en_number\'\n```  \n\n\n## Usage\n### Workflow\n\n\n<p align="center"><img src="docs/static/images/workflow.png"/></p>\n\nExecution flow of UCR is displayed above. Broadly it can be divided into 4 sub-parts:\n\n1. Input (image/folder path or web address) is fed into the <u>Detection</u> model which outputs bounding box coordinates of all the text boxes.\n2. The detected boxes are then checked for <u>Orientation</u> and corrected accordingly.\n3. Next, <u>Recognition</u> model runs on the corrected text boxes. It returns bounding box information and OCR output.\n4. Lastly, an optional <u>Post Processing</u> module is executed to improve/modify the results.\n\n### Quick Start\n\nThe following code snippet shows how to get started with UCR library.\n\n```python\nfrom ucr import UCR\n\n# initialization\nocr = UCR(lang="en_number", device="cpu")\n\n# run prediction\nresult = ocr.predict(\'input_path\', output=\'output_path\')\n\n# for saving annotated image\nresult = ocr.predict(\'input_path\', output=\'output_path\', save_image=True)\n```\nFor complete list of arguments, refer <a href="docs/tldr.md/#argument-list">Argument List</a>\n\n## Model Zoo\n\nA collection of pretrained models for detection, classification and recognition processes is present <a href="ucr.docyard.ai/modelzoo">here</a> !  \nThese models can be useful for out-of-the-box inference on over 25 languages.\n\n\n## Acknowledgement\n\nSubstantial part of the UCR library is either inspired or inherited from the [PaddleOCR](https://github.com/PaddlePaddle/PaddleOCR) library. Wherever possible the repository has been ported from PaddlePaddle to PyTorch framework including the direct translation of model parameters.\nAlso, a big thanks to [Clova AI](https://clova.ai/en/research/research-areas.html), for open sourcing their testing script and pretrained models ([CRAFT](https://github.com/clovaai/CRAFT-pytorch)).  \n\n## License\n\n[Apache License 2.0](LICENSE)\n',
    'author': 'Abhigyan Raman',
    'author_email': 'abhigyan@docyard.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import cv2
import numpy as np
import pyclipper
import torch
from shapely.geometry import Polygon


class DBPostProcess(object):
    """
    The post process for Differentiable Binarization (DB).
    """

    def __init__(
        self,
        thresh=0.3,
        box_thresh=0.7,
        max_candidates=1000,
        unclip_ratio=2.0,
        use_dilation=False,
        **kwargs
    ):
        self.thresh = thresh
        self.box_thresh = box_thresh
        self.max_candidates = max_candidates
        self.unclip_ratio = unclip_ratio
        self.min_size = 3
        self.dilation_kernel = (
            None if not use_dilation else np.array([[1, 1], [1, 1]])
        )

    def boxes_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape
        contours = None

        outs = cv2.findContours(
            (bitmap * 255).astype(np.uint8),
            cv2.RETR_LIST,
            cv2.CHAIN_APPROX_SIMPLE,
        )
        if len(outs) == 3:
            _, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)

        boxes = []
        scores = []
        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            score = self.box_score_fast(pred, points.reshape(-1, 2))
            if self.box_thresh > score:
                continue

            box = self.unclip(points).reshape(-1, 1, 2)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)

            box[:, 0] = np.clip(
                np.round(box[:, 0] / width * dest_width), 0, dest_width
            )
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height
            )
            boxes.append(box.astype(np.int16))
            scores.append(score)
        return np.array(boxes, dtype=np.int16), scores

    def unclip(self, box):
        unclip_ratio = self.unclip_ratio
        poly = Polygon(box)
        distance = poly.area * unclip_ratio / poly.length
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded

    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

        index_1, index_2, index_3, index_4 = 0, 1, 2, 3
        if points[1][1] > points[0][1]:
            index_1 = 0
            index_4 = 1
        else:
            index_1 = 1
            index_4 = 0
        if points[3][1] > points[2][1]:
            index_2 = 2
            index_3 = 3
        else:
            index_2 = 3
            index_3 = 2

        box = [
            points[index_1],
            points[index_2],
            points[index_3],
            points[index_4],
        ]
        return box, min(bounding_box[1])

    def box_score_fast(self, bitmap, _box):
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype(np.int), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype(np.int), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype(np.int), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype(np.int), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype(np.int32), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def __call__(self, outs_dict, shape_list):
        pred = outs_dict["maps"]
        if torch.is_tensor(pred):
            pred = pred.numpy()
        pred = pred[:, 0, :, :]
        segmentation = pred > self.thresh

        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            src_h, src_w, ratio_h, ratio_w = shape_list[batch_index]
            if self.dilation_kernel is not None:
                mask = cv2.dilate(
                    np.array(segmentation[batch_index]).astype(np.uint8),
                    self.dilation_kernel,
                )
            else:
                mask = segmentation[batch_index]
            boxes, scores = self.boxes_from_bitmap(
                pred[batch_index], mask, src_w, src_h
            )

            boxes_batch.append({"points": boxes})

        return boxes_batch



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Modifications copyright (c) 2021 DocYard Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy

from torch.utils.data import DataLoader

__all__ = ["build_dataloader"]


def build_dataloader(config, mode, logger):
    config = copy.deepcopy(config)

    support_dict = ["SimpleDataSet", "LMDBDateSet"]
    module_name = config[mode]["Dataloader"]["name"]
    assert module_name in support_dict, Exception(
        "DataSet only support {}".format(support_dict)
    )
    assert mode in [
        "Train",
        "Eval",
        "Test",
    ], "Mode should be Train, Eval or Test."

    dataset = eval(module_name)(config, mode, logger)
    loader_config = config[mode]["Dataloader"]
    batch_size = loader_config["batch_size_per_card"]
    drop_last = loader_config["drop_last"]
    shuffle = loader_config["shuffle"]
    num_workers = loader_config["num_workers"]

    data_loader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        num_workers=num_workers,
        drop_last=drop_last,
    )

    return data_loader
